/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package es.rickyepoderi.ldappasswd;

import java.net.URI;
import java.net.URISyntaxException;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.Properties;
import javax.naming.Context;
import javax.naming.ldap.ExtendedRequest;
import javax.naming.ldap.InitialLdapContext;
import javax.naming.ldap.LdapContext;
import javax.naming.ldap.StartTlsRequest;
import javax.naming.ldap.StartTlsResponse;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

/**
 * <p>Utility class to perform a ldappasswd against an ldap server. The command
 * is more or less the same than the UNIX counterpart. You can get the usage
 * using the <em>-h</em> option.</p>
 *
 * @author rickyepoderi
 */
public class LdapPasswd {

    private String bindDN = null;
    private String bindPwd = null;
    private String ldapUrl = "ldap://localhost:389";

    private String oldPassword = null;
    private String newPassword = null;
    private String userDN = null;

    private boolean insecure = false;
    private boolean startTls = false;

    public LdapPasswd(String... args) throws URISyntaxException {
        parseArguments(args);
    }

    private static void usage(String error) {
        String nl = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder(nl).append(nl);
        if (error != null && !error.isEmpty()) {
            sb.append("ERROR: ").append(error).append(nl).append(nl);
        }
        sb.append("USAGE: java -jar ldappasswd.jar [-H ldapuri] [-D binddn] [-W] [-w passwd] [-A] [-a oldPasswd] [-S] [-s newPasswd] [-Z] [user]").append(nl)
                .append("  OPTIONS:").append(nl)
                .append("    -H <ldapuri>: DN of the admin user to connect with (default: \"ldap://localhost:389\")").append(nl)
                .append("    -D <binddn>: DN of the admin user to connect with (ex: \"cn=Directory Manager\")").append(nl)
                .append("    -W: Promp for the password of the admin user to connect with").append(nl)
                .append("    -w <passwd>: Password of the admin user to connect with").append(nl)
                .append("    -A: Promp for the old password").append(nl)
                .append("    -a <oldPasswd>: Old password of the user").append(nl)
                .append("    -S: Promp for the new password").append(nl)
                .append("    -s <newPasswd>: New password of the user").append(nl)
                .append("    -k: insecure connection (accepts any certificate)").append(nl)
                .append("    -Z: StartTLS before connecting.").append(nl)
                .append("    -h: Prints this help.").append(nl)
                .append("  ARGUMENTS:").append(nl)
                .append("    [user]: DN of the user to change the password (if none it will be changed to the connected user)").append(nl);
        throw new IllegalArgumentException(sb.toString());
    }

    private String parseStringArgument(String option, String[] args, int i) {
        if (i >= args.length) {
            usage("Option \"" + option + "\" needs a parameter");
        }
        return args[i];
    }

    private void parseArguments(String... args) throws URISyntaxException {
        int i;
        boolean more = true;
        for (i = 0; i < args.length && more; i++) {
            switch(args[i]) {
                case "-H":
                    this.ldapUrl = parseStringArgument("-H", args, ++i);
                    break;
                case "-D":
                    this.bindDN = parseStringArgument("-D", args, ++i);
                    break;
                case "-W":
                    this.bindDN = new String(System.console().readPassword("%s: ", "BindDN password"));
                    break;
                case "-w":
                    this.bindPwd = parseStringArgument("-w", args, ++i);
                    break;
                case "-a":
                    this.oldPassword = parseStringArgument("-A", args, ++i);
                    break;
                case "-A":
                    this.oldPassword = new String(System.console().readPassword("%s: ", "Old password"));
                    break;
                case "-s":
                    this.newPassword = parseStringArgument("-S", args, ++i);
                    break;
                case "-S":
                    String passwd = new String(System.console().readPassword("%s: ", "New password"));
                    this.newPassword = new String(System.console().readPassword("%s: ", "Repeat new password"));
                    if (!passwd.equals(this.newPassword)) {
                        usage("New password does not match!");
                    }
                    break;
                case "-k":
                    this.insecure = true;
                    break;
                case "-Z":
                    this.startTls = true;
                    break;
                case "-h":
                    usage(null);
                    break;
                default:
                    more = false;
                    i--;
            }
        }
        if (i < args.length && args[i].startsWith("-")) {
            usage(String.format("Invalid option \"%s\".", args[i]));
        }
        if (i < args.length) {
            this.userDN = args[i++];
        }
        if (i < args.length) {
            usage("Extra arguments detected!");
        }
        if (startTls && !"ldap".equals(new URI(ldapUrl).getScheme())) {
            usage("StartTLS (-Z option) can only be used in ldap URLs.");
        }
    }

    private void makeInsecure() throws NoSuchAlgorithmException, KeyManagementException {
        TrustManager[] trustAllCerts = new TrustManager[]{
            new X509TrustManager() {
                @Override
                public java.security.cert.X509Certificate[] getAcceptedIssuers() {
                    return null;
                }

                @Override
                public void checkClientTrusted(
                        java.security.cert.X509Certificate[] certs, String authType) {
                }

                @Override
                public void checkServerTrusted(
                        java.security.cert.X509Certificate[] certs, String authType) {
                }
            }
        };
        // Install the all-trusting trust manager globally
        SSLContext sslContext = SSLContext.getInstance("SSL");
        sslContext.init(null, trustAllCerts, new java.security.SecureRandom());
        SSLContext.setDefault(sslContext);
    }

    public PasswordModifyExtendedResponse execute() throws Exception {
        // connect using the connection details
        Properties env = new Properties();
        env.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
        env.put(Context.PROVIDER_URL, ldapUrl);
        env.put("com.sun.jndi.ldap.connect.pool", "false");
        env.put("com.sun.jndi.ldap.read.timeout", "30000");
        env.put("com.sun.jndi.ldap.connect.timeout", "30000");
        if (!startTls && bindDN != null && bindPwd != null) {
            env.put(Context.SECURITY_AUTHENTICATION, "simple");
            env.put(Context.SECURITY_PRINCIPAL, bindDN);
            env.put(Context.SECURITY_CREDENTIALS, bindPwd);
        }

        LdapContext ctx = null;
        System.out.println("# Using LDAP url " + ldapUrl + "...");
        try {
            if (this.insecure) {
                System.out.println("# Using insecure connection...");
                this.makeInsecure();
            }

            if (this.startTls) {
                ctx = new InitialLdapContext(env, null);
                ExtendedRequest tlsRequest = new StartTlsRequest();
                StartTlsResponse tlsResponse = (StartTlsResponse) ctx.extendedOperation(tlsRequest);
                System.out.println("# Performing the StartTLS with user " + bindDN + "... ");
                tlsResponse.negotiate();
                if (bindDN != null && bindPwd != null) {
                    ctx.addToEnvironment(Context.SECURITY_AUTHENTICATION, "simple");
                    ctx.addToEnvironment(Context.SECURITY_PRINCIPAL, bindDN);
                    ctx.addToEnvironment(Context.SECURITY_CREDENTIALS, bindPwd);
                }
            } else {
                System.out.println("# Connecting as user " + bindDN + "... ");
                ctx = new InitialLdapContext(env, null);
            }

            System.out.println("# Performing password change...");
            PasswordModifyExtendedRequest passwordModifyRequest = new PasswordModifyExtendedRequest(
                    this.userDN, this.oldPassword, this.newPassword);
            return (PasswordModifyExtendedResponse) ctx.extendedOperation(passwordModifyRequest);
        } finally {
            if (ctx != null) {
                ctx.close();
            }
        }
    }

    public static void main(String... args) throws Exception {
        LdapPasswd pwdCmd = new LdapPasswd(args);
        PasswordModifyExtendedResponse response = pwdCmd.execute();
        if (response.getGeneratedPassword() != null) {
            System.out.println("# generated passwd: " + response.getGeneratedPassword());
        }
    }
}
