/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package es.rickyepoderi.ldappasswd;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.nio.charset.StandardCharsets;
import javax.naming.NamingException;
import javax.naming.ldap.ExtendedRequest;
import javax.naming.ldap.ExtendedResponse;

/**
 * <p>Implementation for the LDAP Password Modify Extended Operation as a java
 * ExtendedRequest. The idea is not using any extra dependency. Just plain
 * JavaSE.</p>
 *
 * <p>Useful links for other implementations:</p>
 * <ul>
 * <li><a href="https://tools.ietf.org/html/rfc3062">rfc3062</a></li>
 * <li><a href="https://github.com/pingidentity/ldapsdk/blob/master/src/com/unboundid/ldap/sdk/extensions/PasswordModifyExtendedRequest.java">Pingidentity implementation</a></li>
 * <li><a href="https://sourceforge.net/p/jamm/git/ci/master/tree/jamm/src/backend/java/jamm/ldap/ModifyPasswordRequest.java">Simple example</a></li>
 * </ul>
 *
 * @author rickyepoderi
 */
public class PasswordModifyExtendedRequest implements ExtendedRequest {

    /**
     * OID for the LDAP Password Modify Extended Operation.
     */
    public static final String PASSWORD_MODIFY_REQUEST_OID = "1.3.6.1.4.1.4203.1.11.1";

    // constants for the sequence of values in the operation
    protected static final byte UNIVERSAL_SEQUENCE_TYPE = 0x30;
    protected static final byte TYPE_USER_IDENTITY = (byte) 0x80;
    protected static final byte TYPE_OLD_PASSWORD = (byte) 0x81;
    protected static final byte TYPE_NEW_PASSWORD = (byte) 0x82;

    // values for the elements in the operation
    private final String userIdentity;
    private final String oldPassword;
    private final String newPassword;

    /**
     * Constructor using the properties. It seems that the operation admits
     * that all the parameters are null. Although server implementations return
     * errors in general if all are null. For example ApacheDS makes compulsory
     * <em>oldPassword</em> if the user is not an admin and the <em>newPassword</em>
     * cannot be null. Nevertheless 389 allows null <em>newPassword</em> for
     * admins and then the response contains a generated password.
     *
     * @param userIdentity The user DN to perform the password change (null means for the current user)
     * @param oldPassword The old password of the user
     * @param newPassword The new password for the user
     */
    public PasswordModifyExtendedRequest(final String userIdentity, final String oldPassword, final String newPassword) {
        this.userIdentity = userIdentity;
        this.oldPassword = oldPassword;
        this.newPassword = newPassword;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getID() {
        return PASSWORD_MODIFY_REQUEST_OID;
    }

    /**
     * The length is encoded in one byte if less than 127/0x7F. If it's bigger
     * than that it is one byte for the size (0x81..0x84) and the bytes needed
     * to encode the integer (4 bytes max).
     *
     * @param length The length to encode
     * @param buffer The buffer to encode the length to
     */
    static void encodeLengthTo(final int length, final ByteArrayOutputStream buffer) {
        if ((length & 0x7F) == length) {
            buffer.write((byte) length);
        } else if ((length & 0xFF) == length) {
            buffer.write((byte) 0x81);
            buffer.write((byte) (length & 0xFF));
        } else if ((length & 0xFFFF) == length) {
            buffer.write((byte) 0x82);
            buffer.write((byte) ((length >> 8) & 0xFF));
            buffer.write((byte) (length & 0xFF));
        } else if ((length & 0x00FF_FFFF) == length) {
            buffer.write((byte) 0x83);
            buffer.write((byte) ((length >> 16) & 0xFF));
            buffer.write((byte) ((length >> 8) & 0xFF));
            buffer.write((byte) (length & 0xFF));
        } else {
            buffer.write((byte) 0x84);
            buffer.write((byte) ((length >> 24) & 0xFF));
            buffer.write((byte) ((length >> 16) & 0xFF));
            buffer.write((byte) ((length >> 8) & 0xFF));
            buffer.write((byte) (length & 0xFF));
        }
    }

    /**
     * Same than in encode.
     *
     * @param buffer The buffer to decode the length
     * @return The length decoded
     */
    static int decodeLength(final ByteArrayInputStream buffer) {
        int first = buffer.read();
        int length = first & 0x7F;
        if (length != first) {
            final int numLengthBytes = length;
            length = 0;
            for (int i = 0; i < numLengthBytes; i++) {
                length <<= 8;
                length |= (buffer.read() & 0xFF);
            }
        }
        return length;
    }

    /**
     * Encode the string in the buffer in TLV format (UTF-8 is needed).
     *
     * @param type The type of the ASN.1
     * @param str The string to encode
     * @param buffer The buffer to encode to
     */
    static void encodeASN1OctetString(final byte type, final String str, final ByteArrayOutputStream buffer) {
        byte[] value = str.getBytes(StandardCharsets.UTF_8);
        buffer.write(type);
        encodeLengthTo(value.length, buffer);
        buffer.write(value, 0, value.length);
    }

    /**
     * Encodes the previous bytes as a sequence.
     *
     * @param value The previous sequence bytes
     * @return The bytes of the sequence
     */
    static byte[] encodeASN1Sequence(final byte[] value) {
        ByteArrayOutputStream buffer = new ByteArrayOutputStream(value.length + 5);
        buffer.write(UNIVERSAL_SEQUENCE_TYPE);
        encodeLengthTo(value.length, buffer);
        buffer.write(value, 0, value.length);
        return buffer.toByteArray();
    }

    /**
     * Decodes the generated password TLV.
     *
     * @param buffer The byte buffer containing the TLV
     * @return The generated password
     */
    static String decodeGeneratedPassword(final ByteArrayInputStream buffer) throws NamingException {
        if ((byte) buffer.read() != PasswordModifyExtendedResponse.TYPE_GENERATED_PASSWORD) {
            throw new NamingException("Invalid type for generated password returned by the server");
        }
        int length = decodeLength(buffer);
        if (length > 0) {
            byte[] value = new byte[length];
            if (buffer.read(value, 0, length) < length) {
                throw new NamingException("Invalid length");
            }
            return new String(value, StandardCharsets.UTF_8);
        }
        return null;
    }

    /**
     * Decode the sequence with the generated password.
     *
     * @param buffer The byte buffer containing the password
     * @return The generated password
     */
    static String decodeGeneratedPasswordSequence(final ByteArrayInputStream buffer) throws NamingException {
        // check sequence
        if (buffer.read() != UNIVERSAL_SEQUENCE_TYPE) {
            throw new NamingException("Invalid DER sequence returned by the server");
        }
        // parse length
        int length = decodeLength(buffer);
        if (length > 0) {
            return decodeGeneratedPassword(buffer);
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public byte[] getEncodedValue() {
        ByteArrayOutputStream buffer = new ByteArrayOutputStream(128);
        // encode the three elements
        if (userIdentity != null) {
            encodeASN1OctetString(TYPE_USER_IDENTITY, userIdentity, buffer);
        }
        if (oldPassword != null) {
            encodeASN1OctetString(TYPE_OLD_PASSWORD, oldPassword, buffer);
        }
        if (newPassword != null) {
            encodeASN1OctetString(TYPE_NEW_PASSWORD, newPassword, buffer);
        }
        // create the sequence with the bytes
        return encodeASN1Sequence(buffer.toByteArray());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public ExtendedResponse createExtendedResponse(String id, byte[] berValue, int offset, int length) throws NamingException {
        if (length > 0) {
            // decode the password generated
            ByteArrayInputStream input = new ByteArrayInputStream(berValue, offset, length);
            String generatedPassword = decodeGeneratedPasswordSequence(input);
            return new PasswordModifyExtendedResponse(generatedPassword);
        }
        // empty generated password (maybe better return null if nothing?)
        return new PasswordModifyExtendedResponse();
    }
}
